#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include "video_device.h"
#include "hm12.h"
#include "jpeg_dev.h"
#include "YGL2/ytk.h"

#define OUTPUT		stdout

#define PE(s...)	fprintf(stderr, s)
#define PO(s...)	fprintf(OUTPUT, s)

int nmasks;
unsigned char **masks;

int pixel_threshold = 28;
float image_threshold = 0.001;
int frame_id = -3;
struct timeval start_time;

void frame_handler(void *data, void *data_back, void *img_data,
	int width, int height, int csp)
{
	int x, y, v, cy = 0, cu = 0, cv = 0, cy0 = 0;
	unsigned char *d1, *d0, *d1j = NULL, *d1k = NULL, *iy, *iu, *iv;
	int count;
	
	struct tm *hms;
	struct timeval curtime, diff;

	int *mcount;
	int i;
	
	d1 = (unsigned char *)data;
	d0 = (unsigned char *)data_back;
	if ((csp == VIDEO_YV12) || (csp == VIDEO_YU12)) {
		d1j = d1 + (width * height);
		d1k = d1 + (width * height * 5 / 4);
	}

	// img_data is YV12
	iy = (unsigned char *)img_data;
	iv = (unsigned char *)img_data + (width * height);
	iu = (unsigned char *)img_data + (width * height * 5 / 4);
	
	mcount = alloca(nmasks * sizeof(int));
	for (i = 0; i < nmasks; i++) {
		mcount[i] = 0;
	}
	count = 0;
	
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			if (csp == VIDEO_YUYV) {
				cy = d1[x << 1];
				if (!(x & 1)) {
					cu = d1[(x << 1) + 1];
					cv = d1[(x << 1) + 3];
				}
				cy0 = d0[x << 1];
			} else if (csp == VIDEO_YV12) {
				cy = d1[x];
				if (!(x & 1)) {
					cv = d1j[x >> 1];
					cu = d1k[x >> 1];
				}
				cy0 = d0[x];
			} else if (csp == VIDEO_YU12) {
				cy = d1[x];
				if (!(x & 1)) {
					cu = d1j[x >> 1];
					cv = d1k[x >> 1];
				}
				cy0 = d0[x];
			} else if (csp == VIDEO_YUV) {
				cy = d1[x * 3];
				cu = d1[(x * 3) + 1];
				cv = d1[(x * 3) + 2];
				cy0 = d0[x * 3];
			}

			iy[x] = cy;
			iu[x >> 1] = cu;
			iv[x >> 1] = cv;
			
			v = cy - cy0;
			if ((v < -pixel_threshold)
			||  (v > pixel_threshold)) {
				count++;
				iu[x >> 1] = 0;
				iv[x >> 1] = 0;
				for (i = 0; i < nmasks; i++)
				if (masks[i][y * width + x] < 128) {
					mcount[i]++;
					if (i & 1) {
						iu[x >> 1] = 255;
						iv[x >> 1] = 0;
					} else {
						iu[x >> 1] = 0;
						iv[x >> 1] = 255;
					}
				}
			}
		}
		
		// image is YV12
		iy += width;
		if (y & 1) {
			iu += width >> 1;
			iv += width >> 1;
		}
		if (csp == VIDEO_YUYV) {
			d1 += width << 1;
			d0 += width << 1;
		} else if (csp == VIDEO_YUV) {
			d1 += width * 3;
			d0 += width * 3;
		} else if ((csp == VIDEO_YU12) || (csp == VIDEO_YV12)) {
			d1 += width;
			d0 += width;
			if (y & 1) {
				d1j += width >> 1;
				d1k += width >> 1;
			}
		}
	}
	
	gettimeofday(&curtime, NULL);
	
	if (frame_id == 0) {
		PE("Starting analysis...\n");
		start_time = curtime;
	}
	
	timersub(&curtime, &start_time, &diff);
	
	if (frame_id >= 0) {
		hms = localtime(&curtime.tv_sec);

		PO("%10d, %10ld.%03ld, %04d, %02d, %02d, %02d, %02d, "
			"%02d.%03ld, %3d, %d, ",
			frame_id, diff.tv_sec, diff.tv_usec / 1000,
			hms->tm_year + 1900, hms->tm_mon + 1, hms->tm_mday,
			hms->tm_hour, hms->tm_min, hms->tm_sec,
			curtime.tv_usec / 1000,
			hms->tm_yday + 1, ((hms->tm_wday + 6) % 7) + 1);
	
		PO("%6d ", count);
		for (i = 0; i < nmasks; i++) {
			PO("%6d ", mcount[i]);
		}
	
		/* if ((float)count > image_threshold * width * height) {
			PO("MOVE");
		} else {
			PO("____");
		}*/
	
		PO("\n");
		fflush(OUTPUT);
	}
	
	frame_id++;
}


int main(int argc, char **argv)
{
	char *devname;
	video_device dev;
	Workspace img;
	Window win;
	void *capture, *data, *data_back;
	int win_width, win_height;
	int out_fourcc;
	
	int i, w, h;
	
	if (argc < 2) {
		PE("Usage: gemvid <video-device> [mask-file.jpg ...]\n\n"
			"<video-device> is most often /dev/video0\n");
		return(0);
	}
	
	nmasks = argc - 2;
	masks = malloc(nmasks * sizeof(unsigned char *));
	if (!masks) {
		PE("malloc(): %s\n", strerror(errno));
		return(1);
	}
	
	devname = argv[1];
	
	if (video_init(&dev, -1, -1, -1, devname))
		return(1);
	
	if ((dev.fourcc != VIDEO_YV12) && (dev.fourcc != VIDEO_HM12)
	&&  (dev.fourcc != VIDEO_YU12) && (dev.fourcc != VIDEO_YUYV)
	&&  (dev.fourcc != VIDEO_MJPG) && (dev.fourcc != VIDEO_JPEG)) {
		PE("%s: Current pixel format (0x%x) not supported. "
			"Try changing it.\n", dev.device, dev.fourcc);
		video_shutdown(&dev);
		return(1);
	}
	
	for (i = 0; i < nmasks; i++) {
		masks[i] = jpegdev_extract_file(argv[i + 2], &w, &h,
			JPEGDEV_GRAYSCALE);
		if (!masks[i]) {
			PE("Could not open %s\n", argv[i + 2]);
			return(1);
		}
		
		if ((w != dev.width) || (h != dev.height)) {
			PE("%s: %dx%d mask does not match %dx%d video\n",
				argv[i + 2], w, h, dev.width, dev.height);
			return(1);
		}
	}
	
	if (YtkInit()) {
		video_shutdown(&dev);
		return(1);
	}
	
	win = YCreateWindow(0, 0, dev.width, dev.height, 0,
		YGL_FULL_EVENT_MASK);
	if (win == None) {
		YtkClose();
		video_shutdown(&dev);
		return(1);
	}
	YSetName(win, "gemvid", "gemvid");
	
	// YGL2 supports X11 <-> v4l2 colorspace naming conversions
	if (YtkCreateWorkspace(&img, YTK_CSPIMAGE, dev.width, dev.height,
	YTK_YV12)) {
		YDestroyWindow(win);
		YtkClose();
		video_shutdown(&dev);
		return(1);
	}
	
	if (dev.fourcc == VIDEO_HM12) {
		data = malloc(dev.size * 2);
		if (!data) {
			PE("malloc(2x%d): %s\n", dev.size, strerror(errno));
			return(1);
		}
		data_back = data + dev.size;
		out_fourcc = VIDEO_YV12;
	} else if ((dev.fourcc == VIDEO_JPEG) || (dev.fourcc == VIDEO_MJPG)) {
		data = malloc(dev.width * dev.height * 3 * 2);
		if (!data) {
			PE("malloc(%dx%dx3x2): %s\n", dev.width, dev.height,
				strerror(errno));
			return(1);
		}
		data_back = data + dev.width * dev.height * 3;
		out_fourcc = VIDEO_YUV;
	} else {
		data = NULL;
		data_back = malloc(dev.size);
		if (!data_back) {
			PE("malloc(%d): %s\n", dev.size, strerror(errno));
			return(1);
		}
		out_fourcc = dev.fourcc;
	}
	
	while(1) {
		capture = video_capture(&dev);
		if (!capture) return(1);
		
		if (dev.fourcc == VIDEO_HM12) {
			hm12_to_yv12(data, capture, dev.width, dev.height);
		} else if ((dev.fourcc == VIDEO_JPEG)
		|| (dev.fourcc == VIDEO_MJPG)) {
			if (jpegdev_extract(data, capture, dev.size, dev.width,
			dev.height, JPEGDEV_PACKED_YUV)) {
				PE("jpegdev_extract() failed\n");
				// return(1);
			}
		} else {
			data = capture;
		}
		
		frame_handler(data, data_back, img.data,
			dev.width, dev.height, out_fourcc);
		
		memcpy(data_back, data, dev.size);
		
		YGetSize(win, &win_width, &win_height);
		YtkPutExtWorkspace(&img, win, 0, 0, win_width, win_height,
			0, 0, dev.width, dev.height);
		YSync();
		
		// write(1, data, dev.size);
	}
	
	if (dev.fourcc == VIDEO_HM12) free(data);
	free(data_back);
	YtkClose();
	video_shutdown(&dev);
}



